#include "mpi.h"
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>
#include "xmit_file.h"
#include "debug.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif /* HAVE_CONFIG_H */

/* Queue-related */
struct node {
    int val;
    struct node *next;
};

struct queue {
    int size;
    struct node *head, *tail;
};

struct queue *queue_create()
{
    struct queue *q;

    q = malloc(sizeof(struct queue));
    if (q == NULL)
        return NULL;

    q->head = q->tail = NULL;

    q->size = 0;

    return q;
}

int queue_pop(struct queue *q)
{
    pt_debug("q->size = %d\n", q->size);
    if (q->size > 0){
        int val = q->head->val;
        struct node *tmp = q->head;

        q->head = q->head->next;
        if (q->head == NULL)
            q->tail = NULL;
        q->size--;
        pt_debug("q->size = %d\n", q->size);

        free(tmp);

        pt_debug("popping %d\n", val);
        return val;
    } 

    return -1;
}

int queue_push(struct queue *q, int val)
{
    struct node *tmp;

    tmp = malloc(sizeof(struct node));
    if (tmp == NULL)
        return -1;

    tmp->val = val;
    tmp->next = NULL;

    if (q->size == 0){
        q->head = q->tail = tmp;
    } else {
        q->tail->next = tmp;
        q->tail = tmp;
    }

    q->size++;

    return 0;
}

int queue_empty(struct queue *q)
{
    return q->size == 0;
}

/* Arguments and options */
#define MAX_CMD_SIZE	4096
#define MAX_FN_SIZE		1024

int fetch_output = 0;
int fetch_output_to_subdirs = 0;
int retry = 0;
char command[MAX_CMD_SIZE];

void print_usage()
{
    fprintf(stderr, 
            "ptdistrib - run a command on a list of files\n"
            "usage: ptdistrib -frh <command> [files]\n"
            "\t-f\tFetch the output of the command\n"
            "\t-F\tFetch the output of the command into different directories for each host\n"
            "\t-r\tRetry the execution on all files until successful\n"
            "\t-h\tPrint this message\n"
            "\n"
            "<command> should contain one or more instances of {} which will be replaced\n"
            "by a filename\n"
           );
}

int parse_args(int argc, char *argv[])
{
    int c;
    extern int optind;

    if (argc < 3)
        return -1;

    while ((c = getopt(argc, argv, "Ffrh")) != EOF){
        switch (c){
            case 'f':
                fetch_output = 1;
                break;
            case 'F':
                fetch_output_to_subdirs = 1;
            case 'r':
                retry = 1;
                break;
            case 'h':
            case '?':
                return -1;
        }
    }

    strncpy(command, argv[optind++], MAX_CMD_SIZE);

    pt_debug("end of options is %d\n", optind);
    return optind;
}

int replace_brackets(char *cmd, char *filename)
{
    char *prev = cmd, *next;
    char tmp[MAX_CMD_SIZE];
    int cmdlen, fnlen;

    cmdlen = strlen(cmd);
    fnlen = strlen(filename);

    while ((next = strstr(prev, "{}")) != NULL){
        cmdlen += fnlen - 2;
        if (cmdlen >= MAX_CMD_SIZE)
            return -1;
        strcpy(tmp, next+2);
        *next = '\0';
        strcat(cmd, filename);
        strcat(cmd, tmp);
        prev = next + fnlen;	
    }

    return 0;
}

enum {
    FILE_XFER_TAG = 0,
    STATUS_TAG = 10,
    ABORT_TAG = 20,
    CTRL_TAG = 30,
    HOSTNAME_TAG = 40,
    ALL_DONE_TAG = 50
};

enum {
    CTRL_QUIT,
    CTRL_FILE
};

/*D

  ptdistrib - SUT distrib

Description:

ptdistrib distributes a task on a list of files to specified hosts.

Command Line Arguments:

ptdistrib [-m <machine_file>|-M <machine_list>] [-fFrh] <command> [<file1> <file2> <file3>...]

.N machine_args

. -f - fetch the output files of the command being run and place them in the
current directory
. -F - fetch the output files of the command being run and place them in the
separate directories for each different host
. -r - retry (If the processing of a file fails on any host, try to process it again.  Continue until all files have been successfully processed)
. -h - print a usage message
. <command> - the command to run.  All instances of "{}" in the command will be
replaced by the current file being processed
- [<file1> <file2> <file3>...] - the list of files to distribute

Environment Variables:

.N machine_envs

Examples:

To compile all the C files in the current directory into object files and
fetch the object files back to the current host on the machines specified in 
the file "destination_machines" use

.vb
ptdistrib -m destination_machines -f "cc -c {}" *.c
.ve

To do the same action, specifying the machines on the command line, use

.vb
ptdistrib -M "host1 host2 host3" -f "cc -c {}" *.c
.ve

Notes:

Retry should be used sparingly.

 D*/

#define MAX_HOSTNAME_LEN 1024

int main(int argc, char *argv[])
{
    int i;
    int rank, csize;
    int status = 0, global;
    int *statusv = NULL;
    struct queue *avail = NULL, *files = NULL;
    int file_start;
    char hostname[MAX_HOSTNAME_LEN];
    MPI_Status mpistat;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &csize);

    /* Check the arguments */
    file_start = parse_args(argc, argv);
    if (file_start < 0){
        if (rank == 0)
            print_usage();
        MPI_Finalize();
        return 1;
    }

    if (fetch_output_to_subdirs) {
        if (gethostname(hostname, MAX_HOSTNAME_LEN) < 0)
            snprintf(hostname, MAX_HOSTNAME_LEN, "host-%d", rank);
    }

    status = compression_init();

    /* Allocate various structures */
    if (rank == 0){
        pt_debug("Allocating structures\n");
        /* Create the status vector */
        statusv = malloc(sizeof(int)*csize);
        if (statusv == NULL)
            status = -1;

        /* Allocate the available process queue */
        if (status == 0){
            avail = queue_create();
            if (avail == NULL)
                status = -1;
        }

        /* Allocate the file queue */
        if (status == 0){
            files = queue_create();
            if (files == NULL){
                status = -1;
            } else {
                /* initialize the file queue */
                for (i = file_start; i < argc; i++) {
                    pt_debug("adding file %s to queue\n", argv[i]);
                    queue_push(files, i);
                }
            }
        }
    }
    MPI_Allreduce(&status, &global, 1, MPI_INT, MPI_MIN, MPI_COMM_WORLD);
    if (global < 0){
        if (rank == 0)
            perror("Unable to allocate internal structures");
        MPI_Finalize();
        return 1;
    }

    if (rank == 0){
        /* The number of workers working */
        int pending = 0;

        MPI_Gather(&status, 1, MPI_INT, statusv, 1, MPI_INT, 0, 
                MPI_COMM_WORLD);
        for (i = 1; i < csize; i++){
            if (statusv[i] == 0){
                queue_push(avail, i);
            }
            statusv[i] = 0;
        }

        pt_debug("about to enter main loop\n");
        while (!queue_empty(files) || pending){
            if (pending == 0 && queue_empty(avail)){
                fprintf(stderr, "Fatal: Files remain unprocessed, "
                        "but there are no workers available!\n");
                MPI_Finalize();
                return 1;
            }

            while (!queue_empty(avail) && !queue_empty(files)){
                int worker, file, ctrlmsg = CTRL_FILE;

                worker = queue_pop(avail);
                file = queue_pop(files);
                pt_debug("sending %d file #%d, %s\n", worker, file, argv[file]);

                MPI_Send(&ctrlmsg, 1, MPI_INT, worker, CTRL_TAG, MPI_COMM_WORLD);
                status = Send_file(argv[file], worker, FILE_XFER_TAG, MPI_COMM_WORLD);
                switch (status){
                    case XF_FILEERR:
                        fprintf(stderr, "Unable to open %s:", argv[file]);
                        perror("");
                        break;
                    case XF_SUCCESS:
                        pending++;
                        statusv[worker] = file;
                        break;
                    default:
                        pt_debug("file send failed!\n");
                        queue_push(avail, worker);
                        queue_push(files, file);
                        break;
                }
            }

            pt_debug("waiting for a response\n");
            MPI_Recv(&status, 1, MPI_INT, MPI_ANY_SOURCE, STATUS_TAG, 
                    MPI_COMM_WORLD, &mpistat);

            queue_push(avail, mpistat.MPI_SOURCE);
            if (retry && status < 0)
                queue_push(files, statusv[mpistat.MPI_SOURCE]);

            if ((fetch_output || fetch_output_to_subdirs) && status == 0){
                if (fetch_output_to_subdirs) {
                    MPI_Recv(hostname, MAX_HOSTNAME_LEN, MPI_CHAR, 
                            mpistat.MPI_SOURCE, HOSTNAME_TAG, 
                            MPI_COMM_WORLD, &mpistat);
                    if ((mkdir(hostname, 0700) != 0) && (errno != EEXIST)) {
                        fprintf(stderr, "Unable to make directory for %s\n", hostname);
                    } 
                    if (chdir(hostname) != 0) {
                        fprintf(stderr, "Unable to change to directory %s\n", hostname);
                    }
                }
                if (Recv_file_tar(mpistat.MPI_SOURCE, FILE_XFER_TAG, MPI_COMM_WORLD) < 0){
                    if (retry)
                        queue_push(files, statusv[mpistat.MPI_SOURCE]);
                }
                if (fetch_output_to_subdirs) {
                    chdir("..");
                }
            }

            statusv[mpistat.MPI_SOURCE] = 0;
            pending--;
        }

        pt_debug("Done!\n");
        /* Tell everyone to quit */
        while (!queue_empty(avail)){
            int ctrlmsg = CTRL_QUIT;
            int worker = queue_pop(avail);
            MPI_Send(&ctrlmsg, 1, MPI_INT, worker, CTRL_TAG, MPI_COMM_WORLD);
        }

    } else {
        int done = 0, ctrlmsg;
        char tmpdir[L_tmpnam];
        char cmd[MAX_CMD_SIZE];
        char filename[MAX_FN_SIZE];

        /* if we can't make a temporary name, give up */
        if (tmpnam(tmpdir) == NULL)
            status = -1;

        /* if we can't make a temporary directory, give up */
        if ((status == 0) && (mkdir(tmpdir, 0700) != 0))
            status = -1;

        /* if we can't change to the temporary directory, give up */
        if (status == 0 && chdir(tmpdir) != 0)
            status = -1;

        MPI_Gather(&status, 1, MPI_INT, statusv, 1, MPI_INT, 0, 
                MPI_COMM_WORLD);

        if (status < 0){
            MPI_Finalize();
            return 1;
        }

        while (!done) {
            pt_debug("%d receiving control message...\n", rank);
            MPI_Recv(&ctrlmsg, 1, MPI_INT, 0, CTRL_TAG, MPI_COMM_WORLD, &mpistat);

            switch (ctrlmsg){
                case CTRL_QUIT:
                    done = 1;
                    continue;
                case CTRL_FILE:
                    break;
                default:
                    continue;
            }

            pt_debug("%d receiving file...\n", rank);
            if (Recv_file(filename, MAX_FN_SIZE, 0, FILE_XFER_TAG, MPI_COMM_WORLD) < 0){
                pt_debug("%d file receive failed\n", rank);
                continue;
            }

            pt_debug("%d: command is '%s'...\n", rank, command);
            /* Replace the brackets in the command with the filename */
            strcpy(cmd, command);
            if (replace_brackets(cmd, filename) < 0){
                status = -1;
                MPI_Send(&status, 1, MPI_INT, 0, STATUS_TAG, MPI_COMM_WORLD);
                continue;
            }

            pt_debug("Executing '%s'\n", cmd);
            status = system(cmd);
            pt_debug("exit status was %d\n", status);
            MPI_Send(&status, 1, MPI_INT, 0, STATUS_TAG, MPI_COMM_WORLD);

            if (fetch_output || fetch_output_to_subdirs) {
                if (fetch_output_to_subdirs)
                    MPI_Send(hostname, MAX_HOSTNAME_LEN, MPI_CHAR, 0, 
                            HOSTNAME_TAG, MPI_COMM_WORLD);
                Send_file_tar("*", 0, FILE_XFER_TAG, MPI_COMM_WORLD);
            }

            system("rm -rf *");
        }

        rmdir(tmpdir);

    }

    MPI_Finalize();
    return 0;
}
